(* ::Package:: *)

(* Mathematica Package *)

(* Created by the Wolfram Workbench 2010-May-13 *)

(* :Title:Obtuse *)
(* :Context:Obtuse` *)
(* :Author:Ingolf Dahl ingolf.dahl@telia.com*)
(* :Summary: This is 
an package to form obtuse angle shadowing networks in point clouds in 
any number of dimensions and to perform multidimensional distance-based 
interpolation between an arbitrary number of scattered points. *)
(* :Copyright:\[Copyright] 2010 by Ingolf Dahl *)
(* :History: TriangularInterpolate was
	created summer 1993 by Tom Wickham-Jones,
	and published in a package ExtendGraphics described in the book
	Mathematica Graphics: Techniques and Applications.
	Tom Wickham-Jones, TELOS/Springer-Verlag 1994
	- made Native by F. Iannarilli 
	Radial basis function interpolation is translated from MATLAB
	 routines described by William Press in the course "Computational 
	 Statistics with Application to Bioinformatics" and adopted what is 
	 suitable in Mathematica .*)

(* :Package Version:0.0*)
(* :Mathematica Version:7.0*)
(* :Keywords: obtuse angle, shadowing network, 
interpolation, scattered points, n-dimensional, multidimensional
distance matrix, machine learning, connection, neighborhood,
connected points, correlated points, associated point set,
distance-based *)

BeginPackage["Obtuse`",{"Combinatorica`","ComputationalGeometry`"}];


Obtuse::usage="This is an package to form obtuse angle 
shadowing networks in point clouds in any number of dimensions and 
to perform multidimensional distance-based interpolation between 
an arbitrary number of scattered distinct points. 
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/Obtuse\"]\)";

ConnectedFrom::usage="ConnectedTo[j,conlist] gives a list 
of numbers for those points which have connections from point number j 
according to the connection list conlist.
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/ConnectedFrom\"]\) ";

ConnectedNeighbors::usage="ConnectedNeighbors[j,conlist] gives a list 
of numbers for those points which are connected to or from point number j 
according to the connection list conlist.
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/ConnectedNeighbors\"]\) "; 

ConnectedTo::usage="ConnectedTo[j,conlist] gives a list 
of numbers for those points which are connected to point number j 
according to the connection list conlist.
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/ConnectedTo\"]\) "; 

ConnectionList::usage="ConnectionList[distmatrix,(options)] gives 
a list of connected point pairs according to the obtuse-angle-shadowing 
rule in a discrete point set, if distmatrix is the distance matrix for the distinct points. 
If the option Type is set to Directed: The point A in the point set is 
NOT connected to the point B if there exist one other points C in the 
point set,such that C is connected to B and the angles ACB is obtuse. 
Further,two points in the network are NOT connected if the distance 
between them is smaller or equal to SmoothenDistance or 
larger or equal to twice the CutoffRadius. 
If the option Type is set to Undirected: Two points A and B in the point 
set are NOT connected if there exist two other points C and D in the point 
set,such that C is connected to A and D is connected to B and both the angles 
ACB and ADB are obtuse. Further,two points in the network are NOT connected 
if the distance between them is smaller or equal to SmoothenDistance 
or larger or equal to twice the CutoffRadius. Each connected point 
pair occurs only once in the returned list, with the lowest point number 
coming first.
The resulting list is sorted according to the length of the connections.
  \!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/ConnectionList\"]\) "; 

ConnectionsToExclude::usage="ConnectionsToExclude (default {}) is an option to ConnectionList 
and to Interpolation, Method -> \"ObtuseAngle\".  
The connections specified by this option, as a list of point pair rules, will be excluded in 
the ConnectionList algorithm, and the next layer of connected pairs will be returned instead. 
 The list of excluded point pairs does not need to be sorted. 
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/ConnectionsToExclude\"]\) "; 

CutoffRadius::usage="CutoffRadius is an option to ConnectionList. The value Infinity is default. If the distance between two point is larger or equal to 2 times the CutoffRadius value, the two points are not connected.
CutoffRadius is also an option to Interpolation when 
Method->\"ObtuseAngle\" is specified. The interpolated value is not affected by the function value in control points that are more distant than the CutoffRadius value.
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData->\"paclet:Obtuse/ref/CutoffRadius\"]\) "; 

CutoffValue::usage="CutoffValue (default 0.) is an option to Interpolation when 
Method->\"ObtuseAngle\" is specified. This value determines the returned value when all control points are more distant than the CutoffRadius value.
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/CutoffValue\"]\) "; 

DirectedDistanceMatrix::usage="DirectedDistanceMatrix[list] computes the symmetric or asymmetric
  matrix of distance or dissimilarity coefficients between the elements of list.
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/DirectedDistanceMatrix\"]\) "; 

FriendCountList::usage="FriendCountList[conlist] gives a list of the type 
{{1,{1,4}},{2,{2,3,6}},{4,{5,7}}}, with the meaning the point 1 and 4 
are connected to one other point each, the points 2,3 and 6 are connected 
to two points each and the points 5 and 7 are connected to four points each.
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/FriendCountList\"]\) "; 

FriendsFromList::usage="FriendsFromList[conlist] gives a list of list of numbers. 
The sublist FriendsFromList[conlist][[j]] contains the numbers for those points 
which are connected from point number j according to the connection list conlist. 
Thus FriendsFromList[conlist][[j]] should be give the same result as 
ConnectedFrom[j,conlist].
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/FriendsFromList\"]\) "; 

FriendsList::usage="FriendsList[conlist] gives a list of list of numbers. 
The sublist FriendsList[conlist][[j]] contains the numbers for those points 
which are connected from or to point number j according to the connection list conlist. 
Thus FriendsList[conlist][[j]] should be give the same result as 
ConnectedNeighbors[j,conlist].
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/FriendsList\"]\) "; 

FriendsToList::usage="FriendsToList[conlist] gives a list of list of numbers. 
The sublist FriendsToList[conlist][[j]] contains the numbers for those points 
which are connected to point number j according to the connection list conlist. 
Thus FriendsToList[conlist][[j]] should be give the same result as 
ConnectedTo[j,conlist].
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/FriendsFromList\"]\) "; 

InterpolationCoefficients::usage="InterpolationCoefficients[interpolationpoint,controlpointlist,distmat,
conlists,interpolationorder,(options)], 
gives a vector of interpolation coefficients for the point interpolationpoint. 
The scalar product between this vector and the vector of function values in 
the distinct points in the controlpointlist gives the interpolated value in 
the point interpolationpoint. For zeroth order the interpolated value is 
the average of the surrounding points.
The argument distmat should give the distance matrix for the control points. The distance matrix does not need to be symmetric. 
The argument conlists should a list of connection lists with the option 
NeighborLevel running from 1 to interpolationorder. If interpolationorder 
is zero, NeighborLevel should be 1. 
The argument interpolationorder should be 0, 1, 2, or 3.
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/InterpolationCoefficients\"]\) "; 

NeighborLevel::usage="NeighborLevel is an option to ConnectionList. The value 1 is default, and specifies that the nearest neighbor connections should be returned.   The value 2 gives the nearest neighbor connections when the level 1 connections are excluded, and so on.
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/NeighborLevel\"]\) "; 

RadialBasisFunction::usage="RadialBasisFunction is an option to Interpolation when 
Method->\"RBF\" is specified. The value Automatic is default. Automatic means \"multiquadric\" function Sqrt[r^2+r0^2] with r0 as half the median value of all the off-diagonal distances between the control points.
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/RadialBasisFunction\"]\) "; 


ShepardPower::usage="ShepardPower is an option to Interpolation when 
Method->\"Shepard\" is specified. The value 6 is default. The weight of each control point is proportional to the distance, taken to the inverse ShepardPower power. If this value is less than the dimension of the point set plus one, the distant points will dominate the interpolated value.
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/ShepardPower\"]\) "; 

SmoothenDistance::usage="SmoothenDistance is an option to ConnectionList. The value 0. is default. If the distance between two point is smaller or equal than the SmoothenDistance value, the two points are not connected.
SmoothenDistance is also an option to Interpolation when 
Method->\"ObtuseAngle\" is specified. Close to the affected control points, the interpolated value will be an average of the corresponding function values. That implies that the interpolated function will not take exactly the function values in the control points.
\!\(\*ButtonBox[\(More...\),BaseStyle->\"Link\",ButtonData:>\"paclet:Obtuse/ref/SmoothenDistance\"]\) "; 



Begin["`Private`"];

Options[ConnectionList]={ConnectionsToExclude->{},CutoffRadius->Infinity,NeighborLevel->1,SmoothenDistance->0.,Type->Directed};


ConnectionList[distmatrix_,opts___?OptionQ]/;
And[(1==NeighborLevel/. Flatten[{opts}]/. Options[ConnectionList]),
	(Undirected==Type/. Flatten[{opts}]/.Options[ConnectionList])]:=
	Module[{distlist,conlist,friends,codsq,exclude,sd},
codsq=4*(CutoffRadius/. Flatten[{opts}]/.Options[ConnectionList])^2;
exclude=ConnectionsToExclude/. Flatten[{opts}]/.Options[ConnectionList];conlist={};
sd=SmoothenDistance/. Flatten[{opts}]/. Options[ConnectionList];
distlist=Sort[DeleteCases[Flatten[Table[{i,j},{i,1,Length[distmatrix]},
	{j,i+1,Length[distmatrix]}],1],{i_,j_}/;
	Or[distmatrix[[i,j]]<=sd,distmatrix[[j,i]]<=sd,
		And[distmatrix[[i,j]]>=codsq,distmatrix[[j,i]]>=codsq],
		MemberQ[exclude,(i->j)],MemberQ[exclude,(j->i)]]],
		(Min[distmatrix[[Sequence@@#1]],
			distmatrix[[Sequence@@Reverse[#1]]]]<Min[distmatrix[[Sequence@@#2]],
			distmatrix[[Sequence@@Reverse[#2]]]]&)];
friends=Table[{},{Length[distmatrix]}];
Do[If[Or[Catch[Do[If[distmatrix[[distlist[[i,1]],
	friends[[distlist[[i,1]],j]]]]+distmatrix[[friends[[distlist[[i,1]],j]],
	distlist[[i,2]]]]<distmatrix[[distlist[[i,1]],distlist[[i,2]]]],Throw[False]],
		{j,1,Length[friends[[distlist[[i,1]]]]]}];
		Throw[True]],
		Catch[Do[If[distmatrix[[distlist[[i,2]],
			friends[[distlist[[i,2]],j]]]]+distmatrix[[friends[[distlist[[i,2]],j]],
			distlist[[i,1]]]]<distmatrix[[distlist[[i,2]],distlist[[i,1]]]],Throw[False]],
				{j,1,Length[friends[[distlist[[i,2]]]]]}];
				Throw[True]]],
				friends[[distlist[[i,1]]]]=Append[friends[[distlist[[i,1]]]],distlist[[i,2]]];
				friends[[distlist[[i,2]]]]=Append[friends[[distlist[[i,2]]]],distlist[[i,1]]];
conlist={conlist,Rule[distlist[[i,1]],distlist[[i,2]]]}],{i,Length[distlist]}];
Flatten[conlist]];


ConnectionList[distmatrix_,opts___?OptionQ]:=Module[{nl,exclude},
	exclude=ConnectionsToExclude/. Flatten[{opts}]/.Options[ConnectionList];
	Nest[{#,ConnectionList[distmatrix,ConnectionsToExclude->#,NeighborLevel->1,opts]}&@
		Join[#[[1]],#[[2]]]&,{exclude,ConnectionList[distmatrix,
			ConnectionsToExclude->exclude,NeighborLevel->1,opts]},nl-1][[2]]/;
			(1<(nl=NeighborLevel/. Flatten[{opts}]/. {NeighborLevel->1}))];


ConnectionList[distmatrix_,opts___?OptionQ]/;
And[(1==NeighborLevel/. Flatten[{opts}]/. Options[ConnectionList]),
	(Directed==Type/. Flatten[{opts}]/.Options[ConnectionList])]:=
	Module[{distlist,conlist,codsq,exclude,sd,friendpointers,connumber,friend,j},
codsq=4*(CutoffRadius/. Flatten[{opts}]/.Options[ConnectionList])^2;
exclude=ConnectionsToExclude/. Flatten[{opts}]/. Options[ConnectionList];
sd=(SmoothenDistance/. Flatten[{opts}]/. Options[ConnectionList])^2;
conlist={};connumber=Length[distmatrix];friendpointers[i_]:=0;
distlist=Sort[DeleteCases[Flatten[Table[{i,j},{i,1,Length[distmatrix]},
	{j,1,Length[distmatrix]}],1],{i_,j_}/;Or[i==j,distmatrix[[i,j]]<=sd,distmatrix[[i,j]]>=codsq,
	MemberQ[exclude,(i->j)]]],(distmatrix[[Sequence@@#1]]<distmatrix[[Sequence@@#2]]&)];
Do[j=distlist[[i,2]];If[Catch[While[friendpointers[j]=!=  0,friend=friendpointers[j][[1]];j=friendpointers[j][[2]];
If[distmatrix[[distlist[[i,1]],friend]]+distmatrix[[friend,distlist[[i,2]]]]<distmatrix[[distlist[[i,1]],distlist[[i,2]]]],Throw[False]]];Throw[True]],connumber+=1;
friendpointers[j]={distlist[[i,1]],connumber};
conlist={conlist,Rule[distlist[[i,1]],distlist[[i,2]]]}],{i,Length[distlist]}];
Flatten[conlist]];

Options[DirectedDistanceMatrix] = {DistanceFunction -> Automatic};

ddf[a_List, b_List] := 
  If[Or[MatchQ[a, {(True | False) ..}], 
    MatchQ[b, {(True | False) ..}]], JaccardDissimilarity[a, b], 
   Plus @@ ((ddf @@ #) & /@ Transpose[{a, b}])];
ddf[a_?NumericQ, b_] := (a - b)^2;
ddf[a_, b_?NumericQ] := (a - b)^2;
ddf[a_String, b_] := EditDistance[a, b];
ddf[a_, b_String] := EditDistance[a, b];
ddf[a_ /; Or[a, Not[a]], b_] := JaccardDissimilarity[a, b];
ddf[a_, b_ /; Or[b, Not[b]]] := JaccardDissimilarity[a, b];

DirectedDistanceMatrix[pts_List, opts___?OptionQ] := 
  Module[{distfunc}, 
   distfunc = 
    DistanceFunction /. Flatten[{opts}] /. 
     Options[DirectedDistanceMatrix];
   If[distfunc === Automatic, distfunc=ddf];     
    Outer[distfunc, pts, pts, 1]];

Options[InterpolationCoefficients]={DistanceFunction->Automatic,CutoffRadius->Infinity,Type->Directed};

f[x_, 1, -1] := Piecewise[{{1, x <= 1/2}}, 0];

fcor[x_, -1] := Piecewise[{{1, x < 1}}, 0];

f[x_, 1, 0] := Piecewise[{{1, x < 1}}, 0];

fcor[x_, 0] := Piecewise[{{1, x < 1}}, 0];

f[x_, 1, 1] := Piecewise[{{1, x <= 0}, {1 - x, x <= 1}}, 0];

fcor[x_, 1] := Piecewise[{{1, x <= 0.5}, {2 (1 - x), x < 1}}, 0];

f[x_, 1, 2] := 
  Piecewise[{{1 - Exp[2*x] + Exp[x], x <= 0}, {1 - x, x <= 1}}, 
   Exp[2*(1 - x)] - Exp[1 - x]];
   
f[x_, 2, 2] := 
  Piecewise[{{1, x <= 0}, {1 - 3 x^2 + 2 x^3, x <= 1}}, 0];
  
fcor[x_, 2] := 
  Piecewise[{{1, x <= 0.5}, {-4 + 24 x - 36 x^2 + 16 x^3, x < 1}}, 0];
  
f[x_, 1, 3] := 
  Piecewise[{{1 + 5*Exp[x]/2 - 4*Exp[2*x] + 3*Exp[3*x]/2, 
     x <= 0}, {1 - x, x <= 1}}, -5*Exp[1 - x]/2 + 4*Exp[2*(1 - x)] - 
    3*Exp[3*(1 - x)]/2];
    
f[x_, 2, 3] := 
  Piecewise[{{1 - 3*Exp[x] + 6*Exp[2*x] - 3*Exp[3*x], 
     x <= 0}, {1 - 3 x^2 + 2 x^3, x <= 1}}, 
   3*Exp[1 - x] - 6*Exp[2*(1 - x)] + 3*Exp[3*(1 - x)]];
   
f[x_, 3, 3] := 
  Piecewise[{{1, x <= 0}, {1 - 10 x^3 + 15 x^4 - 6 x^5, x <= 1}}, 0];
  
fcor[x_, 3] := 
  Piecewise[{{1, 
     x <= 0.5}, {32 - 240 x + 720 x^2 - 1040 x^3 + 720 x^4 - 192 x^5, 
     x < 1}}, 0];

InterpolationCoefficients[interpolationpoint_, controlpointlist_, 
    distmat_, conlist_, interpolationorder_, 
    opts___?OptionQ] /; (Length[conlist] >= interpolationorder) := 
  Module[{list1, ilist, x, distfunc, cor, undir, 
    conlistlocal}, 
   distfunc = 
    DistanceFunction /. Flatten[{opts}] /. 
     Options[InterpolationCoefficients];
   If[distfunc === Automatic, 
    distfunc = 
     ddf];
   undir = (Undirected == Type /. Flatten[{opts}] /. 
      Options[InterpolationCoefficients]);
   undir = TrueQ[undir];
   cor = CutoffRadius /. Flatten[{opts}] /. 
     Options[InterpolationCoefficients];
   If[undir, 
    conlistlocal = 
     conlist /. {(Rule[i_, j_]) :> Sequence[Rule[i, j], Rule[j, i]]}, 
    conlistlocal := conlist];
   ilist = Table[1, {i, Length[controlpointlist] + 1}];
   list1 = 
    Table[distfunc[interpolationpoint, controlpointlist[[j]]], {j, 1, 
       Length[controlpointlist]}] /. {a1_Rational | a1_Integer :> 
       1.*a1};
   If[cor == Infinity, ilist[[-1]] = 0, 
    Do[Piecewise[{{Null,ilist[[i]] == 0.}}, 
      ilist[[i]] = 
       ilist[[i]]*fcor[Sqrt[list1[[i]]]/cor, interpolationorder]; 
      ilist[[-1]] = 
       ilist[[-1]]*(1 - 
          fcor[Sqrt[list1[[i]]]/cor, interpolationorder])], {i, 
      Length[ilist] - 1}]];
   Do[Do[Piecewise[{{Null,ilist[[conlistlocal[[m, i, 2]]]] == 0.}}, 
      x = (distmat[[conlistlocal[[m, i, 1]], 
           conlistlocal[[m, i, 2]]]] - 
          list1[[conlistlocal[[m, i, 1]]]] + 
          list1[[conlistlocal[[m, i, 2]]]])/(2*
          distmat[[conlistlocal[[m, i, 1]], conlistlocal[[m, i, 2]]]]);
      ilist[[conlistlocal[[m, i, 2]]]] = 
       ilist[[conlistlocal[[m, i, 2]]]]*
        f[x, m, interpolationorder]], {i, 
      Length[conlistlocal[[m]]]}], {m, Max[1, interpolationorder], 
     1, -1}];
x = Total[ilist];
Piecewise[{{ilist,x == 0.}}, ilist = ilist/x]];


ConnectedNeighbors[j_,conlist_]:=Union[Cases[conlist,(a_->j):>a],Cases[conlist,(j->a_):>a]];


ConnectedTo[j_,conlist_]:=Union[Cases[conlist,(a_->j):>a]];


ConnectedFrom[j_,conlist_]:=Union[Cases[conlist,(j->a_):>a]];


FriendsList[conlist_]:=Module[{maxconlist,fl},maxconlist=Max[List @@@conlist];
fl=Table[{},{maxconlist}];
Do[AppendTo[fl[[conlist[[i,1]]]],conlist[[i,2]]];
	AppendTo[fl[[conlist[[i,2]]]],conlist[[i,1]]],{i,Length[conlist]}];Map[Union,fl]];


FriendsToList[conlist_]:=Module[{maxconlist,fl},maxconlist=Max[List @@@conlist];
fl=Table[{},{maxconlist}];
Do[AppendTo[fl[[conlist[[i,2]]]],conlist[[i,1]]],{i,Length[conlist]}];Map[Union,fl]];


FriendsFromList[conlist_]:=Module[{maxconlist,fl},maxconlist=Max[List @@@conlist];
fl=Table[{},{maxconlist}];
Do[AppendTo[fl[[conlist[[i,1]]]],conlist[[i,2]]],{i,Length[conlist]}];Map[Union,fl]];


FriendCountList[conlist_]:=Module[{friendnumbers,countlist},
	friendnumbers=Length/@ FriendsList[conlist];
	countlist=
	Append[#,{}]&/@List/@Union[friendnumbers];
Do[countlist=countlist/.{{friendnumbers[[i]],a1_List}:>{friendnumbers[[i]],
	Append[a1,i]}},{i,Length[friendnumbers]}];countlist];

TriangularInterpolateNative[
  pnts_List /; 
   MatrixQ[N[pnts[[All, 1]]], NumberQ] && Length[First[pnts]] === 2, 
  opts___Rule] :=
 TriangularInterpolateNative[N[pnts], 
  DelaunayTriangulation[N[pnts[[All, 1]]], Hull -> True], opts];

TriangularInterpolateNative[
  pnts_List /; 
   MatrixQ[N[pnts[[All, 1]]], NumberQ] && 
    Length[First[pnts]] === 2, {adj_, hull_}, opts___Rule] := 
 Block[{t, tri, num, work, minx, maxx, miny, maxy, tris, 
   ext},(*adjacency list->triangles,courtesy dh in MathGroup:
  "If we partition the neighbours into all possible pairs and add
      the vertex, we obtain all triangles that contain this vertex.
      However, this gives the same triangle several times.
      Therefore, we sort the triangle points and use
      Unique to get rid of the multiplicity."*)
  t = Function[x, Prepend[#, x[[1]]] & /@ Partition[x[[2]], 2, 1]] /@ 
    adj;
  tri = Sort /@ Flatten[t, 1] // Union;
  num = Range[Length[tri]];
  tris = Map[Part[pnts, #] &, tri];
  work = Map[Min[Part[Transpose[Part[Transpose[#], 1]], 1]] &, tris];
  minx = SortFun[work];
  work = Map[Max[Part[Transpose[Part[Transpose[#], 1]], 1]] &, tris];
  maxx = SortFun[work];
  work = Map[Min[Part[Transpose[Part[Transpose[#], 1]], 2]] &, tris];
  miny = SortFun[work];
  work = Map[Max[Part[Transpose[Part[Transpose[#], 1]], 2]] &, tris];
  maxy = SortFun[work];
  TriangularInterpolating[{hull, tri, pnts, minx, maxx, miny, maxy}]];


SortFun[ data_] :=
    	Sort[
	    		Transpose[ { Range[ Length[ data]], data}],
	    		Less[ Part[ #1, 2], Part[#2, 2]]&]

Format[ t_TriangularInterpolating] := "TriangularInterpolating[ <> ]";

FindFirstMin[ x_, imin_, imax_, list_] :=
	Block[ {pos},
		Which[
			x < Part[ list, 1, 2], {},
			x >=  Part[ list, -1, 2], Map[ First, list],
			True,
				pos = FindFirstMin1[ x, imin, 
imax, list] ;
				If[  x < Part[ list, pos, 2] ||
					(pos < Length[ list] && 
x >=  Part[ list, pos+1,2]),
					Print[ "Error in 
Bounding Box calc"]] ;
				Map[ First, Take[ list, pos]]]
		];

FindFirstMin1[ x_, imin_, imax_, list_] :=
	Block[{pos},
		If[ imin === imax,
			imin,
			pos = Floor[ (imin+imax)/2] ;
			If[ x < Part[ list, pos, 2],
				FindFirstMin1[ x, imin, pos, 
list],
				If[ x < Part[ list, pos+1, 2],
					pos,
					FindFirstMin1[ x, pos+1, 
imax, list]
					]
				]
			]
		];

FindFirstMax[ x_, imin_, imax_, list_] :=
	Block[ {pos},
		Which[
			x <= Part[ list, 1, 2], Map[ First, list],
			x >  Part[ list, -1, 2], {},
			True,
				pos = FindFirstMax1[ x, imin, 
imax, list] ;
				If[  x <= Part[ list, pos, 2] ||
					(pos < Length[ list] && 
x >  Part[ list, pos+1,2]),
					Print[ "Error in 
Bounding Box calc"]] ;
				Map[ First, Drop[ list, pos]]]
		];

FindFirstMax1[ x_, imin_, imax_, list_] :=
	Block[{pos},
		If[ imin === imax,
			imin,
			pos = Floor[ (imin+imax)/2] ;
			If[ x <= Part[ list, pos, 2],
				FindFirstMax1[ x, imin, pos, 
list],
				If[ x <= Part[ list, pos+1, 2],
					pos,
					FindFirstMax1[ x, pos+1, 
imax, list]
					]
				]
			]
		];

PointInTri[{x_, y_}, {{x1_, y1_}, {x2_, y2_}, {x3_, y3_}}] := 
 Block[{t1, t2, t3, eps = 5 $MachineEpsilon}, 
  t1 = -x1 y + x2 y + x y1 - x2 y1 - x y2 + x1 y2;
  t2 = -x2 y + x3 y + x y2 - x3 y2 - x y3 + x2 y3;
  t3 = x1 y - x3 y - x y1 + x3 y1 + x y3 - x1 y3;
  If[t1 > -eps && t2 > -eps && t3 > -eps || 
    t1 < eps && t2 < eps && t3 < eps, True, False]];

FindTri[pt_, tri_, pts_, rng_] := 
 Block[{tst}, tst = Part[pts, Part[tri, First[rng]]][[All, 1]];
  If[PointInTri[pt, tst], First[rng], 
   If[Length[rng] === 1, {}, FindTri[pt, tri, pts, Rest[rng]]]]];

InterpolatePointInTri[{x_, 
   y_}, {{{x1_, y1_}, z1_}, {{x2_, y2_}, z2_}, {{x3_, y3_}, z3_}}] := 
 Block[{den, a, b, c}, 
  den = x2 y1 - x3 y1 - x1 y2 + x3 y2 + x1 y3 - x2 y3;
  a = -(y2 z1) + y3 z1 + y1 z2 - y3 z2 - y1 z3 + y2 z3;
  b = x2 z1 - x3 z1 - x1 z2 + x3 z2 + x1 z3 - x2 z3;
  c = x3 y2 z1 - x2 y3 z1 - x3 y1 z2 + x1 y3 z2 + x2 y1 z3 - x1 y2 z3;
  a/den x + b/den y + c/den];



TriangularInterpolating[{hull_, tri_, pts_, minx_, maxx_, miny_, 
    maxy_}][{x_?NumberQ, y_?NumberQ}] := 
 Module[{x0, y0, x1, y1, x2, y2, triang, len, pnts2, normals, 
   bisectors, posang, j, bispoint1, bispoint2, dist1, dist2, rng, dirang}, 
  len = Length[minx];
  x0 = FindFirstMin[x, 1, len, minx];(*In x0 and above*)
  x1 = FindFirstMax[x, 1, len, maxx];(*In x1 and below*)
  y0 = FindFirstMin[y, 1, len, miny];
  y1 = FindFirstMax[y, 1, len, maxy];
  rng = Intersection[x0, x1, y0, y1];
  If[rng =!= {}, triang = FindTri[{x, y}, tri, pts, rng], triang = {}];
  If[triang === {}, (pnts2 = pts[[All, 1]];
    normals = 
     Table[Mod[
       ArcTan @@ (pnts2[[hull[[Mod[i + 1, Length[hull], 1]]]]] - 
           pnts2[[hull[[i]]]]) - \[Pi]/2, 2 \[Pi]], {i, Length[hull]}];
    bisectors = 
     Table[Mod[
       normals[[Mod[i - 1, Length[hull], 1]]] + 
        1/2 Mod[normals[[i]] - normals[[Mod[i - 1, Length[hull], 1]]],
           2 \[Pi]], 2 \[Pi]], {i, Length[normals]}]; 
           
    dirang = 
     Table[
       ArcTan @@ ({x, y} - pnts2[[hull[[i]]]]), {i, Length[hull]}];
       
    posang = 
     Table[Mod[
       dirang[[i]] - bisectors[[i]], 
       2 \[Pi], -\[Pi]], {i, Length[hull]}];
      
    j = Length[hull]; i = 1; 
    While[i < j && (Mod[dirang[[i]] - dirang[[i+1]],2*Pi,-Pi]<0 || posang[[i]] <= 0 || posang[[i + 1]] > 0), 
     i = i + 1]; j = i;
    bispoint1 = ({x2, y2} /. 
        Solve[{{Cos[normals[[j]]], Sin[normals[[j]]]}.{x2 - x, 
             y2 - y} == 
           0, {-Sin[bisectors[[j]]], 
             Cos[bisectors[[j]]]}.{x2 - pnts2[[hull[[j]]]][[1]], 
             y2 - pnts2[[hull[[j]]]][[2]]} == 0}, {x2, y2}])[[1]];
    bispoint2 = ({x2, y2} /. 
        Solve[{{Cos[normals[[j]]], Sin[normals[[j]]]}.{x2 - x, 
             y2 - y} == 
           0, {-Sin[bisectors[[Mod[j + 1, Length[hull], 1]]]], 
             Cos[bisectors[[Mod[j + 1, Length[hull], 1]]]]}.{x2 - 
              pnts2[[hull[[Mod[j + 1, Length[hull], 1]]]]][[1]], 
             y2 - pnts2[[hull[[Mod[j + 1, Length[hull], 1]]]]][[
               2]]} == 0}, {x2, y2}])[[1]];
    dist1 = EuclideanDistance[{x, y}, bispoint1];
    dist2 = EuclideanDistance[{x, y}, bispoint2];
    dist2/(dist1 + dist2)*pts[[hull[[j]]]][[2]] +
     dist1/(dist1 + dist2)*
      pts[[hull[[Mod[j + 1, Length[hull], 1]]]]][[2]])
   (*else*), triang = Part[pts, Part[tri, triang]];
   InterpolatePointInTri[{x, y}, triang]]];

End[];

Protect[Evaluate[Context[]<>"*"]];
EndPackage[];


Unprotect[Interpolation];
Interpolation[funcvaluelist_,opts___?OptionQ]:=
Module[{pts, fvalues,dm,io,cl,cor,corval,nl,tl},
cor=CutoffRadius/. Flatten[{opts}]/.Options[Interpolation];
corval=CutoffValue/. Flatten[{opts}]/.Options[Interpolation]/. 
	{CutoffValue->0*funcvaluelist[[1,2]]};
io=InterpolationOrder/. Flatten[{opts}]/. {InterpolationOrder->1};
If[funcvaluelist=={},Return[corval]];
pts=Transpose[funcvaluelist][[1]];
fvalues=Append[Transpose[funcvaluelist][[2]],corval];
dm=DirectedDistanceMatrix[pts,opts];
cl=ConnectionList[dm,opts];
(Which[io==-1,(InterpolationCoefficients[#,pts,dm,{cl},-1,opts].fvalues)&,
	io==0,(InterpolationCoefficients[#,pts,dm,{cl},0,opts].fvalues)&,
	io==1,(InterpolationCoefficients[#,pts,dm,{cl},1,opts].fvalues)&,
io==2,nl=ConnectionList[dm,ConnectionsToExclude->cl,opts];
(InterpolationCoefficients[#,pts,dm,{cl,nl},2,opts].fvalues)&,
io==3,nl=ConnectionList[dm,ConnectionsToExclude->cl,opts];
tl=ConnectionList[dm,ConnectionsToExclude->Join[cl,nl],opts];
(InterpolationCoefficients[#,pts,dm,{cl,nl,tl},3,opts].fvalues)&])
]/;And["ObtuseAngle"===(Method /.Flatten[List[opts]]),MemberQ[{-1,0,1,2,3},InterpolationOrder/.
	 Flatten[{opts}]/. {InterpolationOrder->1}]];
	 
Interpolation[funcvaluelist_,x_,opts___?OptionQ]:=
	 Interpolation[funcvaluelist,opts][x]/;And["ObtuseAngle"===(Method /.Flatten[List[opts]]),
	 	MemberQ[{0,1,2,3},InterpolationOrder/. Flatten[{opts}]/. {InterpolationOrder->1}]];

Interpolation[funcvaluelist_,x_,opts___?OptionQ]:=Module[{pts, fvalues, dp, phi, distfunc}, 
   dp = ShepardPower /. Flatten[{opts}] /. Options[Interpolation];
   distfunc = 
    DistanceFunction /. Flatten[{opts}] /. 
      Options[Interpolation] /. {DistanceFunction -> 
       Automatic}; 
       If[distfunc === Automatic, 
 distfunc = Obtuse`Private`ddf];
       If[funcvaluelist == {}, Return[0.]];
   {pts, fvalues} = Transpose[funcvaluelist];
   phi = (distfunc[x, #] & /@ pts) /. {a1_Rational | a1_Integer :> 1.*a1} ;
   If[Or @@ ((0 == #) & /@ phi), phi = If[# == 0, 1, 0] & /@ phi, 
    phi = (#^(-dp/2)) & /@ phi];
   (phi.fvalues)/Total[phi]] /; 
  "Shepard" === (Method /. Flatten[List[opts]]);
Interpolation[funcvaluelist_,opts___?OptionQ]:=Interpolation[funcvaluelist,#,opts]&/;"Shepard"===(Method /.Flatten[List[opts]]);

Interpolation[funcvaluelist_, x_, opts___?OptionQ] := 
  Module[{dlist, minval, distfunc},
    distfunc = 
     DistanceFunction /. Flatten[{opts}] /. 
       Options[Interpolation] /. {DistanceFunction -> Automatic};
    If[distfunc === Automatic, 
     distfunc = 
      Obtuse`Private`ddf];
    If[funcvaluelist == {}, Return[0.]];
    dlist = {distfunc[x, #[[1]]], #[[2]]} & /@ funcvaluelist;
    minval = dlist[[1]];
    If[#[[1]] < minval[[1]], minval = #] & /@ dlist; minval[[2]]] /;
    "Voronoi" === (Method /. Flatten[List[opts]]);
Interpolation[funcvaluelist_, opts___?OptionQ] := 
  Interpolation[funcvaluelist, #, opts] & /; 
   "Voronoi" === (Method /. Flatten[List[opts]]);

Interpolation[funcvaluelist_,opts___?OptionQ]:=Module[{pts, fvalues,distfunc,radbasfunc,phimat,wgts,len,distmat,r02},
distfunc=DistanceFunction/. Flatten[{opts}]/.Options[Interpolation]/.
	{DistanceFunction->Automatic};
	If[distfunc == Automatic, 
 distfunc = Obtuse`Private`ddf];
radbasfunc=RadialBasisFunction/. Flatten[{opts}]/.Options[Interpolation];
If[funcvaluelist=={},Return[0.]];
{pts,fvalues}=Transpose[funcvaluelist];
distmat=DirectedDistanceMatrix[pts,opts] /.{a1_Rational | a1_Integer :> 1.*a1};
If[radbasfunc==Automatic,r02=Median[Flatten[Table[Drop[distmat[[i]], {i}], {i, Length[pts]}]]];radbasfunc=((Sqrt[#+r02/4])&)];
phimat =Map[radbasfunc,distmat,{2}];
wgts=LinearSolve[phimat,fvalues];len=Length [pts];
Sum[wgts[[i]]*(radbasfunc @distfunc [#,pts[[i]]]),{i,len}]&]/;
"RBF"===(Method /.Flatten[List[opts]]);

Interpolation[funcvaluelist_,x_,opts___?OptionQ]:=
Interpolation[funcvaluelist,opts][x]/;"RBF"===(Method /.Flatten[List[opts]]);

Interpolation[funcvaluelist : {{{_, _}, _} ..}, opts___?OptionQ] := 
  Obtuse`Private`TriangularInterpolateNative[funcvaluelist] /; ("Delaunay" === (Method /. 
       Flatten[List[opts]]));

Options[Interpolation]=Union[Options[Interpolation],
	{DistanceFunction->Automatic,CutoffRadius->Infinity,
		RadialBasisFunction->Automatic,ShepardPower->6,Type->Directed}];
Protect[Interpolation];
